unit UnitMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, DGLOpenGL, glBitmap;

type
  TForm1 = class(TForm)
    procedure FormCreate(Sender: TObject);
    procedure FormResize(Sender: TObject);

    procedure EngineCreate;
    procedure EngineDestroy;
    procedure LoadShaders;
    procedure IdleHandler(Sender: TObject; var Done: Boolean);
    function LoadFragmentandVertexShader(FFilename: string; VFilename: string): GLHandleARB;
    procedure CheckFrameBufferObject;

    procedure CreateFBO;

    procedure FormDestroy(Sender: TObject);
  private
    StartTime, TimeCount, FrameCount  : Cardinal; //FrameCounter
    Frames, DrawTime                  : Cardinal; //& Timebased Movement

    procedure Render;
    procedure DrawObjects;
    procedure SetupGL;
    procedure Errorhandler;
  public
    DC                                : HDC;  //Handle auf Zeichenflche
    RC                                : HGLRC;//Rendering Context
  end;

var
  Form1: TForm1;

  //Texturen
  ShadowShader,FinalShadowsShader: GLHandleARB;

  shadow_map,
  shadow_fbo,
  shadow_size : TGLuint;

implementation

{$R *.dfm}

function glSlang_GetInfoLog(pShader: GLHandleARB): String;  //Shaderfehlerausgabe
var
  blen, slen: GLInt;
  InfoLog: PGLCharARB;
begin
  glGetShaderiv(pShader, GL_INFO_LOG_LENGTH , @blen);    //Lnge des Logs
  if blen > 1 then   //Wenn etwas drinsteht, dann wird der Wert als Rckgabewert ausgegeben
  begin
    GetMem(InfoLog, blen * SizeOf(GLCharARB));
    glGetShaderInfoLog(pShader, blen, slen, InfoLog);
    Result :=InfoLog;
    Dispose(InfoLog);
  end;
end;

procedure TForm1.CheckFrameBufferObject;
var
  error: GlEnum;
begin
  error := glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
  case error of
    GL_FRAMEBUFFER_COMPLETE_EXT:
      Exit;
    GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
      raise Exception.Create('Incomplete attachment');
    GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
      raise Exception.Create('Missing attachment');
    GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
      raise Exception.Create('Incomplete dimensions');
    GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
      raise Exception.Create('Incomplete formats');
    GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
      raise Exception.Create('Incomplete draw buffer');
    GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
      raise Exception.Create('Incomplete read buffer');
    GL_FRAMEBUFFER_UNSUPPORTED_EXT:
      raise Exception.Create('Framebufferobjects unsupported');
  end;
end;

function TForm1.LoadFragmentandVertexShader;
var
  ProgramObject, FragmentShaderObject, VertexShaderObject: GLHandleARB;
  FShader: TStringList;
  FShaderText: AnsiString;
  FShaderLength: Integer;
  VShader: TStringList;
  VShaderText: AnsiString;
  VShaderLength: Integer;
begin
  if FileExists(FFilename)= True then
  if FileExists(VFilename)= True then
  begin
   ProgramObject:= glCreateProgramObjectARB; //Programmobjekt erzeugen

   FragmentShaderObject:= glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB); //Vertex-/Fragmentshader erzeugen
   VertexShaderObject:= glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);

   FShader:= TStringList.Create;      //Shadercode laden, in ANSISTRING speichern und die Lnge speichern
   FShader.LoadFromFile(FFilename);
   FShaderText:= FShader.Text;
   FShaderLength:= Length(FShaderText);
   FShader.Free;

   VShader:= TStringList.Create;  //Shadercode laden, in ANSISTRING speichern und die Lnge speichern
   VShader.LoadFromFile(VFilename);
   VShaderText:= VShader.Text;
   VShaderLength:= Length(VShaderText);
   VShader.Free;

   glShaderSourceARB(VertexShaderObject, 1, @VShaderText, @VShaderLength); //ShaderCode mit Shaderobjekt verbinden
   glShaderSourceARB(FragmentShaderObject, 1, @FShaderText, @FShaderLength);

   glCompileShaderARB(FragmentShaderObject); //Shader kompilieren und eventuelle Fehlermeldung herausgeben
   if glSlang_GetInfoLog(FragmentShaderObject)<>'' then MessageDlg('FragmentShader: '+glSlang_GetInfoLog(FragmentShaderObject),mtError,[mbOK],0);
   glCompileShaderARB(VertexShaderObject);
   if glSlang_GetInfoLog(VertexShaderObject)<>'' then MessageDlg('VertexShader: '+glSlang_GetInfoLog(VertexShaderObject),mtError,[mbOK],0);

   glAttachObjectARB(ProgramObject, FragmentShaderObject); //Shaderobjekte mit Programmobjekt verbinden
   glAttachObjectARB(ProgramObject, VertexShaderObject);


   glDeleteObjectARB(FragmentShaderObject);   //Shaderobjekte lschen
   glDeleteObjectARB(VertexShaderObject);
   glLinkProgramARB(ProgramObject);    //Shaderprogramm linken

   Result:= ProgramObject;
  end
  else
   MessageDlg('Shaderdateien existieren nicht.',mtError,[mbOK],0);
end;

procedure TForm1.CreateFBO;
begin
  shadow_size := 4096; // Auflsung der Schatten

  glGenTextures(1, @shadow_map);
  glBindTexture(GL_TEXTURE_2D, shadow_map);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT16, shadow_size, shadow_size, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, NIL);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

  glGenFramebuffersEXT(1, @shadow_fbo);
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, shadow_fbo);
  glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, shadow_map, 0);
  glDrawBuffer (GL_FALSE);
  glReadBuffer(GL_FALSE);
  CheckFrameBufferObject;  //FBO prfen
end;

procedure TForm1.DrawObjects;  //Rendern der Objekte
const
    light_position1 : Array[0..3] of GlFloat = (10.0, 10.0, 0.0, 1.0);
    light_position2 : Array[0..3] of GlFloat = (-10.0, 10.0, 0.0, 1.0);
    light_ambient  : Array[0..3] of GlFloat = (0.8, 0.8, 0.8, 1.0);
    light_diffuse  : Array[0..3] of GlFloat = (0.8, 0.8, 0.8, 1.0);

begin
  glLightfv(GL_LIGHT0, GL_AMBIENT,  @light_ambient[0]);
  glLightfv(GL_LIGHT0, GL_DIFFUSE,  @light_diffuse[0]);
  glLightfv(GL_LIGHT0, GL_POSITION, @light_position1[0]);
  glLightfv(GL_LIGHT1, GL_AMBIENT,  @light_ambient[0]);
  glLightfv(GL_LIGHT1, GL_DIFFUSE,  @light_diffuse[0]);
  glLightfv(GL_LIGHT1, GL_POSITION, @light_position2[0]);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT1);

  glLoadIdentity;
  glTranslatef(20,-25,-100);
  glRotatef(45,0,1,0);
  glPushMatrix;
	glColor4f(0.3,0.3,0.3,1);
	glBegin(GL_QUADS);
	glVertex3f(-35,2,-35);
	glVertex3f(-35,2, 15);
	glVertex3f( 15,2, 15);
	glVertex3f( 15,2,-35);
	glEnd();
  glPopMatrix;

	glColor4f(0.9,0.9,0.9,1);
  glTranslatef(0,0,-5);
  glPushMatrix;
  glBegin(GL_QUADS);
    glVertex3f(0,10,0);
    glVertex3f(0,0,0);
    glVertex3f(10,0,0);
    glVertex3f(10,10,0);
  glend;
  glPopMatrix;
  //-------
end;


procedure TForm1.EngineCreate;
begin
  DC:=GetDC(Handle);
  if not InitOpenGL then Application.Terminate;
  RC:=CreateRenderingContext(DC,[opDoubleBuffered],32,24,0,0,0,0);
  ActivateRenderingContext(DC,RC);
  SetupGL;
  LoadShaders;
  Application.OnIdle:=Form1.IdleHandler;
end;

procedure TForm1.EngineDestroy;
begin
  DeactivateRenderingContext;
  DestroyRenderingContext(RC);
  ReleaseDC(Handle,DC);
end;

procedure TForm1.Errorhandler;
begin
  Form1.Caption:=gluErrorString(glGetError);
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  EngineCreate;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  EngineDestroy;
end;

procedure TForm1.FormResize(Sender: TObject);
var tmpBool: Boolean;
begin
  glViewport(0,0,ClientWidth,ClientHeight);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;
  gluPerspective(45.0,ClientWidth/ClientHeight,1,1000);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;
  IdleHandler(Sender, tmpBool);
end;

procedure TForm1.IdleHandler(Sender: TObject; var Done: Boolean);
var light,lights: integer;
begin
  StartTime:=GetTickCount;
  //---------------------
  //Rendern
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;
  gluPerspective(45.0,ClientWidth/ClientHeight,1,1000);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;


  //Shader fr Dualparabolische Maps aktivieren
  glUseProgramObjectARB(ShadowShader);

  //Framebufferobjekt aktivieren
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, shadow_fbo);
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);

  lights:=2;
  for light := 0 to lights-1 do //zwei Lichter
  begin
    glViewport(round(shadow_size/2*(lights/4)), round(shadow_size/4 *(light mod 4)), round(shadow_size/4), round(shadow_size/4));
    glUniform1iARB(glGetUniformLocationARB(ShadowShader, 'light'), light);
    glUniform1iARB(glGetUniformLocationARB(ShadowShader, 'renderpass'), 0);
    DrawObjects;
    glViewport(round(shadow_size/4+shadow_size/2*(lights/4)), round(shadow_size/4*(light mod 4)), round(shadow_size/4), round(shadow_size/4));
    glUniform1iARB(glGetUniformLocationARB(ShadowShader, 'renderpass'),1);
    DrawObjects;
  end;

  // Framebufferobjekt deaktivieren
  glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

  //Richtigen Shader aktivieren
  glUseProgramObjectARB(FinalShadowsShader);
  //Viewport an Fenstergre anpassen
  glViewport(0, 0, ClientWidth, ClientHeight);

  //Frame rendern
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
  DrawObjects;

  SwapBuffers(dc);
  //---------------------

  DrawTime:=GetTickCount-StartTime;
  inc(TimeCount,DrawTime);
  inc(FrameCount);

  if TimeCount>1000 then
  begin
    Frames:=FrameCount;
    TimeCount:=TimeCount-1000;
    FrameCount:=0;
    Caption:=IntToStr(Frames)+' FPS';
    //Errorhandler;
  end;

  Done:=False;
end;
procedure TForm1.LoadShaders;
begin
  ShadowShader:=LoadFragmentandVertexShader('Shader\Shadow\parabol.frag','Shader\Shadow\parabol.vert');
  FinalShadowsShader:=LoadFragmentandVertexShader('Shader\Shadow\final.frag','Shader\Shadow\final.vert');
end;

procedure TForm1.Render; //Wird nicht benutzt
begin
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;
  gluPerspective(45.0,ClientWidth/ClientHeight,1,1000);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;
  glTranslatef(0,-5,-20);

  DrawObjects;

  SwapBuffers(dc);
end;


procedure TForm1.SetupGL;
begin
  glClearColor(0.3,0.4,0.7,0);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
  glEnable(GL_TEXTURE_2D);
end;

end.
